import mimetypes
import signal
import sys
import time
import typing
from collections import defaultdict
from hashlib import md5
from pathlib import Path
from string import Template
from urllib.parse import quote_plus

from bottle import (
    Bottle,
    HTTPError,
    HTTPResponse,
    redirect,
    request,
    response,
    static_file,
)
from morebuiltins.functools import lru_cache_ttl
from morebuiltins.utils import get_hash, is_running, read_size, read_time, ttime
from psutil import Process

from ..config import Config as MConfig
from ..core import JobDir, PythonDir, Taska, VenvDir, WorkspaceDir
from .console_template import console_template

app = Bottle()
keepalive_timeout = 60
keepalives = {}
# import sys

# sys.path.append("../../")
# from taska.core import DirBase
logger = MConfig.init_logger()


class Config:
    pwd = ""
    salt = md5(Path(__file__).read_bytes()).hexdigest()
    root_path = Path.cwd()
    # file size limit
    max_file_size = 1024 * 100
    console_template = Template(console_template)


class AuthPlugin(object):
    # avoid tries too many times
    blacklist: typing.Dict[str, int] = defaultdict(lambda: 0)
    cookie_max_age = 7 * 86400

    def get_sign(self, now: int, ip: str):
        _hash = md5(
            f"{now+self.cookie_max_age}{ip}{Config.salt}{Config.pwd}".encode()
        ).hexdigest()
        return f"{now+self.cookie_max_age}{_hash}"

    @lru_cache_ttl(ttl=300, maxsize=100, controls=True)
    def get_params_s(self, rule: str):
        return md5(f"{rule}{Config.salt}".encode()).hexdigest()

    def check_blacklist(self, client_ip, now):
        b = self.blacklist[client_ip]
        if b:
            if b > now:
                self.blacklist[client_ip] = b + 5
                timeleft = self.blacklist[client_ip] - now
                raise HTTPError(429, f"Too many tries, retry at {timeleft}s later.")
            else:
                self.blacklist.pop(client_ip, None)

    @lru_cache_ttl(ttl=3600 * 1, maxsize=1000, controls=True)
    def check_cookie(self, sign, client_ip, now):
        if sign:
            try:
                then = int(sign[:10])
                if now > then:
                    return False
                else:
                    return sign == self.get_sign(then - self.cookie_max_age, client_ip)
            except ValueError:
                return False
        else:
            return False

    def is_valid(self, rule):
        client_ip = request.environ.get("HTTP_X_FORWARDED_FOR") or request.environ.get(
            "REMOTE_ADDR"
        )
        if not client_ip:
            raise HTTPError(401, "No client ip")
        now = int(time.time())
        if self.check_blacklist(client_ip, now):
            return True
        sign = request.cookies.get("sign")
        cookie_ok = self.check_cookie(sign, client_ip, now)
        if rule == "/login":
            if request.method == "GET":
                request.environ["cookie_ok"] = cookie_ok
                return True
            else:
                # POST
                self.handle_post_pwd(rule, client_ip, cookie_ok, now)
        s = self.get_params_s(rule)
        if cookie_ok:
            response.set_header("s", s)
            return True
        else:
            # params s auth
            params = request.params
            s_valid = "s" in params and params["s"] == s
            return s_valid

    def handle_post_pwd(self, rule, client_ip, cookie_ok, now):
        pwd = request.forms.get("pwd")
        if not pwd:
            raise HTTPError(401, "No password?")
        if cookie_ok or not Config.pwd:
            # modify pwd
            Config.pwd = pwd
            self.check_cookie.cache.clear()
        if pwd == Config.pwd:
            # correct password
            from_url = request.cookies.get("from_url")
            res = response.copy(cls=HTTPResponse)
            res.status = 303
            res.set_cookie(
                "sign",
                self.get_sign(now, client_ip),
                path="/",
                max_age=self.cookie_max_age * 0.95,
            )
            if from_url and rule != "/login":
                res.delete_cookie("from_url")
            res.body = ""
            res.set_header("Location", from_url or "/")
            raise res
        else:
            # wrong password
            self.blacklist[client_ip] = now + 5
            raise HTTPError(401, "Invalid password")

    def apply(self, callback, context):
        rule = context["rule"]

        def wrapper(*args, **kwargs):
            valid = self.is_valid(rule)
            if not valid:
                if rule not in {"/login", "/favicon.ico"}:
                    response.set_cookie("from_url", request.url, path="/", max_age=3600)
                redirect("/login")
            request.environ["auth_ok"] = 1
            res = callback(*args, **kwargs)
            return res

        return wrapper


@app.get("/")
def index():
    # 1. console
    # 2. view
    return r"""
<style>
button{
    list-style-type: none;
    display: inline-block;
    vertical-align: middle;
    zoom: 1;
    border: 1px solid #d4d4d4;
    font-weight: bold;
    font-family: "Helvetica Neue Light", "Helvetica Neue", Helvetica, Arial, "Lucida Grande", sans-serif;
    margin: 20% 50px 20% 50px;
    text-decoration: none;
    text-align: center;
    border-radius: 60%;
    box-shadow: inset 0px 1px 1px rgba(255, 255, 255, 0.5), 0px 1px 2px rgba(0, 0, 0, 0.2);
    width: 300px;
    line-height: 30%;
    height: 30%;
    padding: 0px;
    border-width: 4px;
    font-size: 3em;
    background: -webkit-linear-gradient(top, #00b5e5, #008db2);
    background-color: #00a1cb;
    border-color: #007998;
    color: white;
    text-shadow: 0 -1px 1px rgba(0, 40, 50, 0.35);
    cursor: pointer;
}
button:hover{
    background: -webkit-linear-gradient(top, #dcdcdc, #189086);
}
</style>
<body style='width: 50%;height: 100%;margin: 0 auto;'>
<a href='/view//'><button>Files</button></a>
<a href='/console'><button>Console</button></a>
</body>
"""


@app.get("/favicon.ico")
def favico():
    response.body = '<svg xmlns="http://www.w3.org/2000/svg"  viewBox="0 0 128 128" width="64px" height="64px"><path fill="#F7F7FB" d="M64 9A55 55 0 1 0 64 119A55 55 0 1 0 64 9Z"/><path fill="#DEDFE6" d="M64,9C33.6,9,9,33.6,9,64s24.6,55,55,55s55-24.6,55-55S94.4,9,64,9z M64,105.2c-22.8,0-41.2-18.5-41.2-41.2S41.2,22.8,64,22.8s41.2,18.5,41.2,41.2S86.8,105.2,64,105.2z"/><path fill="#D8D7D5" d="M64 59.4A4.6 4.6 0 1 0 64 68.6A4.6 4.6 0 1 0 64 59.4Z"/><path fill="#464C55" d="M64,122C32,122,6,96,6,64S32,6,64,6s58,26,58,58S96,122,64,122z M64,12c-28.7,0-52,23.3-52,52s23.3,52,52,52s52-23.3,52-52S92.7,12,64,12z"/><path fill="#464C55" d="M64.1,67.1c-0.8,0-1.5-0.3-2.1-0.9s-0.9-1.3-0.9-2.1L61,36.5c0-1.7,1.3-3,3-3l0,0c1.7,0,3,1.3,3,3l0.1,24.6L82.3,61l0,0c1.6,0,3,1.3,3,3c0,1.7-1.3,3-3,3L64.1,67.1L64.1,67.1z"/><path fill="#464C55" d="M64,71.6c-4.2,0-7.6-3.4-7.6-7.6s3.4-7.6,7.6-7.6s7.6,3.4,7.6,7.6S68.2,71.6,64,71.6z M64,62.4c-0.9,0-1.6,0.7-1.6,1.6c0,0.9,0.7,1.6,1.6,1.6c0.9,0,1.6-0.7,1.6-1.6S64.9,62.4,64,62.4z"/></svg>'
    _md5 = get_hash(response.body)
    response.headers["Cache-Control"] = "public, max-age=%s" % 3600 * 24
    response.add_header("ETag", _md5)
    response.add_header("Content-Type", "image/svg+xml")
    return response


@app.get("/login")
@app.post("/login")
def login():
    if request.method == "GET":
        if Config.pwd and not request.environ.get("cookie_ok"):
            placeholder = "Input the password"
        else:
            placeholder = "Reset the password"
        return r"""
    <form style="width: 100%;height: 100%;" action="/login" method="post">
    <input autofocus style="text-align: center;font-size: 5em;width: 100%;height: 100%;" type="password" name="pwd" placeholder="{placeholder}">
    </form>""".format(placeholder=placeholder)


@app.get("/init/<dir_type>")
def init(dir_type):
    referer = request.query["referer"]
    name = request.query.get("name")
    root = Config.root_path
    target_dir = root.joinpath(referer).resolve()
    if not (target_dir.exists() and target_dir.is_relative_to(root)):
        return "path not found"
    if dir_type == "requirements":
        return VenvDir.ensure_pip_install(target_dir)
    c = {i.__name__: i for i in [JobDir, PythonDir, VenvDir, WorkspaceDir]}[dir_type]
    if c is PythonDir:
        python = request.query.get("python", "")
        c.prepare_dir(target_dir, name, python=python)
    else:
        c.prepare_dir(target_dir, name)
    referer = referer or "/"
    return redirect(f"/view/{referer}")


def get_list_html(path: Path):
    html = "<a style='color: black' href='/'>Home</a> - "
    parts = path.relative_to(Config.root_path.parent).parts
    for index, part in enumerate(parts):
        if index == 0:
            html += f" <a style='color:blue' href='/view//'>{part}</a> /"
        else:
            p = "/".join(parts[1 : index + 1])
            html += f" <a style='color:blue' href='/view/{p}'>{part}</a> /"
    html = html.rstrip("/")
    path_arg = "/".join(parts[1:])
    if JobDir.is_valid(path) or JobDir.is_valid(path.parent):
        html += "<hr>"
        kill_html = ""
        for pid_dir in [path.parent, path]:
            try:
                if pid_dir.is_dir():
                    pid = int(pid_dir.joinpath("pid.txt").read_bytes())
                    if is_running(pid):
                        kill_html = f" | <a style='color:red' href='/console?kill={pid}&signal=15'>Kill - {pid}</a>"
            except (FileNotFoundError, ValueError):
                pass
        html += f"<a style='color:red' href='/launch/{path_arg}?timeout=2'>Launch Job</a> | <a style='color:#009879' href='/console'>Console</a>{kill_html}"
    elif path.is_dir():
        if path == Config.root_path:
            "create python dir"
            html += f" | <form style='color:red' method='get' action='/init/{PythonDir.__name__}'><input placeholder='dir name' name='name'> <input placeholder='python_path' name='python'><input style='display:none' name='referer' value='{path_arg}' name='python'><input type='submit' value='Create PythonDir'></form>"
        elif PythonDir.is_valid(path):
            "create venv dir"
            html += f" | <form style='color:red' method='get' action='/init/{VenvDir.__name__}'><input placeholder='dir name' name='name'><input style='display:none' name='referer' value='{path_arg}'><input type='submit' value='Create VenvDir'></form>"
        elif VenvDir.is_valid(path):
            "fresh requirements"
            html += f" | <a style='color:red' target='_blank' href='/init/requirements?referer={path_arg}'>PIP install</a>"
        elif VenvDir.is_valid(path.parent) and path.name == "workspaces":
            "create workspace dir"
            html += f" | <form style='color:red' method='get' action='/init/{WorkspaceDir.__name__}'><input placeholder='dir name' name='name'><input style='display:none' name='referer' value='{path_arg}'><input type='submit' value='Create WorkspaceDir'></form>"
        elif WorkspaceDir.is_valid(path.parent) and path.name == "jobs":
            "create job dir"
            html += f" | <form style='color:red' method='get' action='/init/{JobDir.__name__}'><input placeholder='dir name' name='name'><input style='display:none' name='referer' value='{path_arg}'><input type='submit' value='Create JobDir'></form>"
    elif path.is_file() and path.suffix == ".py" and WorkspaceDir.is_valid(path.parent):
        html += f" | <form style='color:red' method='get' action='/init/{JobDir.__name__}'><input placeholder='dir name' name='name'><input style='display:none' name='referer' value='{path_arg}'><input type='submit' value='Create Job'></form>"
    html += "<hr>"
    text_arg = ""
    file_name_arg = ""
    old_color = "#696969"
    new_color = "#00c308"
    now = time.time()
    if path.is_dir():
        path_list = sorted(
            path.iterdir(), key=lambda i: f"-{i.name}" if i.is_dir() else i.name
        )
        for _path in path_list:
            try:
                p = _path.relative_to(Config.root_path).as_posix()
                mtime = _path.stat().st_mtime
                if _path.is_dir():
                    color = "darkorange"
                    icon = "&#128194;"
                    size = " - "
                    stat_color = old_color
                else:
                    color = "black"
                    icon = "&#128196;"
                    size = read_size(_path.stat().st_size, 1, shorten=True)
                    if now - mtime < 5 * 60:
                        stat_color = new_color
                    else:
                        stat_color = old_color
                time_stat = f"{ttime(mtime)}({read_time(now-mtime, shorten=True):->8})"
                stat = f"<span style='color:{stat_color};width:260px;display: inline-block;font-size: 0.8em'> | {time_stat} | {size}</span>"
                file_url = f"{request.url.rstrip('/')}/{quote_plus(_path.name)}"
                dir_disabled = " disabled" if _path.is_dir() else ""
                rename_html = f"<form action='/rename' method='get' style='display: inline;'><input style='display:none' name='old_path' value='{path_arg}/{_path.name}'><input type='text' name='name' value='{_path.name}'><input type='submit' value='Rename'></form> | "
                html += f"{rename_html}<button onclick='delete_path(`{file_url}?action=delete`)'>Delete</button> | <a href='{file_url}?action=download'><button{dir_disabled}>Download</button></a> | <a href='{file_url}?action=view'><button{dir_disabled}>View</button></a> {stat} <a style='color:{color}' href='/view/{p}'>{icon} {_path.name}</a>"
                # add rename form
                html += "<br>"
            except FileNotFoundError:
                continue
    else:
        file_name_arg = path.name
        p = path.relative_to(Config.root_path).as_posix()
        mtime = path.stat().st_mtime
        if now - mtime < 5 * 60:
            stat_color = new_color
        else:
            stat_color = old_color
        _path = path
        time_stat = f"{ttime(mtime)}({read_time(now-mtime, shorten=True):->8})"
        stat = f"<span style='color:{stat_color};font-size: 0.8em;width:260px;display: inline-block;'> | {read_size(path.stat().st_size, 1)}|{time_stat}</span>"
        html += f"<button onclick='delete_path(`{request.url}?action=delete`)'>Delete</button> | <a href='{request.url}?action=download'><button>Download</button></a> | <a href='{request.url}?action=view'><button>View</button></a> {stat} <br>"
        if path.stat().st_size < Config.max_file_size:
            text_arg = path.read_bytes().decode("utf-8", "replace")
    max_text_tip = f"preview text-only file_size < {read_size(Config.max_file_size)}"
    html += """<hr><form action="/upload" method="post" enctype="multipart/form-data" id="upload_form">
<input type="hidden" name="path" value="{path_arg}">
File Name:
<input type="text" name="file_name" value="{file_name_arg}"> or <input type="file" name="upload_file"><br>
<textarea placeholder="{max_text_tip}" title="{max_text_tip}" id="text" name="text" style='width:100%;height:50%;border: groove;padding: 2em;font-size: 1.5em;text-wrap: pretty;'>{text_arg}</textarea>
<br>
<input style="font-size: 1.5em;" type="submit" value="Upload <Ctrl+Enter>" /></form>
""".format(
        path_arg=path_arg,
        file_name_arg=file_name_arg,
        text_arg=text_arg,
        max_text_tip=max_text_tip,
    )
    html += r"""<script>
document.addEventListener('DOMContentLoaded', function() {
    var upload_form = document.getElementById('upload_form');
    upload_form.addEventListener('keydown', function(event) {
        if (event.ctrlKey && event.key === 'Enter') {
            event.preventDefault();
            document.getElementById('upload_form').submit();
        }
    });
});
</script>"""
    delete_code = r"""<script>function delete_path(url){
    var isConfirmed = confirm('Are you sure you want to delete this item?');
    if (isConfirmed) {
        window.location.href = url;
    }
    }</script>"""
    return f"<body style='width:80%;margin: 0 auto;'>{html}{delete_code}</body>"


@app.get("/rename")
def rename():
    old_path = request.query["old_path"].lstrip("/")
    path = Config.root_path.joinpath(old_path).resolve()
    if path.exists() and path.is_relative_to(Config.root_path):
        path.rename(path.with_name(request.query["name"]))
    redirect(f"/view/{'/'.join(old_path.split('/')[:-1]) or '/'}")


@app.route("/launch/<path:path>")
def launch(path):
    root = Config.root_path
    _path: Path = root.joinpath(path).resolve()
    if not (_path.exists() and _path.is_relative_to(root)):
        return "path not found"
    timeout = int(request.query.get("timeout", 0))
    job_dir = Taska.launch_job(_path, timeout)
    parts = job_dir.relative_to(Config.root_path.parent).parts
    path_arg = "/".join(parts[1:])
    return redirect(f"/view/{path_arg}")


@app.get("/view")
@app.get("/view/")
def redirect_view_root():
    redirect("/view//")


@app.get("/view/<path:path>")
def list_dir(path):
    if path == "/":
        path = ""
    root = Config.root_path
    real_path: Path = root.joinpath(path).resolve()
    action = request.query.get("action")
    if not (real_path.exists() and real_path.is_relative_to(root)):
        return "path not found"
    elif action == "delete":
        if not real_path.parent.is_relative_to(root):
            return "path not found"
        if real_path.is_dir():
            Taska.safe_rm_dir(real_path)
        else:
            real_path.unlink()
        back = "/".join(request.path.split("/")[:-1])
        if back == "/view":
            return redirect("/view//")
        redirect(back)
    elif action == "download":
        if not real_path.exists():
            return HTTPError(400, "path not found")
        elif not real_path.is_relative_to(root):
            return HTTPError(400, "bad path")
        elif real_path.is_dir():
            return HTTPError(400, "not support download dir")
        else:
            content_type = "application/octet-stream"
            file_content = static_file(
                real_path.as_posix(), real_path.parent.as_posix(), content_type
            )
            response.headers["Content-Disposition"] = (
                f'attachment; filename="{real_path.name}"'
            )
            response.body = file_content
            return response
    elif action == "view":
        grep = request.query.get("grep")
        if real_path.is_file():
            if grep:
                encoding = request.query.get("encoding") or "utf-8"
                with open(real_path, "r", encoding=encoding) as f:
                    lines = []
                    for line in f:
                        if grep in line:
                            lines.append(line)
                    result = "".join(lines)
                    return f"<pre>{result}</pre>"
            else:
                ct = mimetypes.guess_type(real_path.as_posix())
                if ct[0]:
                    response.content_type = ct[0]
                return real_path.read_bytes()
        else:
            return "not a file"
    elif "tail" in request.query:
        return handle_tail(real_path, get_hash((time.time(), real_path.as_posix())))

    else:
        return get_list_html(real_path)


def handle_tail(path: Path, event_id):
    if not path.is_file():
        raise ValueError("not a file")
    tail = int(request.query["tail"])
    encoding = request.query.get("encoding", "utf-8")
    interval = int(request.query.get("interval", 1))
    with open(path, "r", encoding=encoding) as f:
        if tail:
            for index, line in enumerate(f):
                pass
            min_index = index + 1 - tail
            f.seek(0)
            yield "<pre style='font-size: 1.5em;'>"
            for index, line in enumerate(f):
                if index >= min_index:
                    yield line
            yield "</pre>"
        else:
            keepalives[event_id] = int(time.time() + keepalive_timeout)
            yield (
                "<script> (function () { setInterval(() => document.readyState !== 'complete' && fetch('/keepalive?e=%s', { method: 'HEAD' }), %s); })()</script>"
                % (event_id, (keepalive_timeout * 1000 // 2))
            )
            yield "<pre style='font-size: 1.5em;'>"
            # tail -F
            # end of file
            f.seek(path.stat().st_size)
            while True:
                line = f.readline()
                if line:
                    yield line
                else:
                    if time.time() > keepalives.get(event_id, 0):
                        keepalives.pop(event_id, None)
                        break
                    elif path.stat().st_size < f.tell():
                        f.seek(0)
                    time.sleep(interval)
            yield "</pre>"


@app.route("/keepalive", method="HEAD")
def keepalive():
    if request.query.get("d"):
        keepalives.pop(request.query["e"], None)
    else:
        now = int(time.time())
        keepalives[request.query["e"]] = now + keepalive_timeout
    for k, v in list(keepalives.items()):
        if now > v:
            keepalives.pop(k, None)


@app.post("/upload")
def upload():
    file_name = request.forms.get("file_name")
    upload_file = request.files.get("upload_file")
    text = request.forms.get("text")
    path = request.forms.get("path")
    target_dir = Config.root_path.joinpath(request.forms.get("path"))
    if target_dir.is_file() and target_dir.name == file_name:
        target_dir = target_dir.parent
    if not target_dir.is_dir() or not target_dir.is_relative_to(Config.root_path):
        return HTTPError(400, "bad path")
    if upload_file.raw_filename:
        file_name = file_name or upload_file.raw_filename
        upload_file.save(
            target_dir.joinpath(file_name).resolve().as_posix(),
            overwrite=True,
        )
        upload_file.file.close()
    else:
        if not file_name:
            return HTTPError(400, "file_name must be set if text is not null")
        target_file = target_dir.joinpath(file_name).resolve()
        if not target_file.is_relative_to(Config.root_path):
            return HTTPError(400, "bad path")
        if file_name.endswith("/"):
            target_file.mkdir(parents=True, exist_ok=True)
        else:
            target_file.parent.mkdir(parents=True, exist_ok=True)
            target_file.write_text(text, encoding="utf-8", newline="")
    redirect(f"/view/{path}")


@app.get("/console")
def console():
    root = Config.root_path
    kill = request.query.get("kill")
    if kill:
        pid = int(kill)
        if root.joinpath(f"pids/{pid}").is_file():
            signal = int(request.query.get("signal") or 2)
            proc = Process(pid)
            if proc.is_running():
                if signal == 9:
                    proc.kill()
                elif signal == 15:
                    proc.terminate()
                elif signal == 2:
                    if sys.platform == "win32":
                        proc.kill()
                    else:
                        proc.send_signal(2)
                else:
                    raise ValueError("bad signal")
                proc.wait(5)
        redirect(request.headers.get("Referer") or "/console")
    pids_dir = root.joinpath("pids")
    pids = []
    for pid_path in root.rglob("pid.txt"):
        pid = None
        try:
            pid = int(pid_path.read_bytes())
        except FileNotFoundError:
            pass
        finally:
            if pid:
                if is_running(pid):
                    pids.append(pid)
                else:
                    pids_dir.joinpath(str(pid)).unlink(missing_ok=True)
                    pid_path.unlink(missing_ok=True)
    m_file = root.joinpath("max_workers")
    if m_file.is_file():
        max_workers = m_file.read_text().strip() or "-"
    else:
        max_workers = "-"
    items: dict = Taska.get_pids_info(pids)
    Taska.LATEST_PROC_CACHE.update(items)
    # [{'pid': 10916, 'status': 'running', 'job_dir': 'default/venv1/workspaces/workspace1/jobs/job1', 'start_at': '2024-08-05 21:36:35', 'elapsed': '19 secs', 'memory': '17 MB'}]
    th_list = [
        f"<th>{k}</th>"
        for k in [
            f"*/{max_workers} - <a style='color: #ffffff' href='/'>Home</a>",
            "pid",
            "status",
            "start_at",
            "end_at",
            "elapsed",
            "memory",
            "job_dir",
            "kill-2",
            "kill-15",
            "kill-9",
        ]
    ]
    tr_list = []
    rows = sorted(
        Taska.LATEST_PROC_CACHE.items(),
        key=lambda x: x[1]["start_at"],
        reverse=True,
    )
    cache_length = Taska.CACHE_LENGTH
    for row_id, (pid, item) in enumerate(rows, 1):
        if pid in items:
            tr_list.append(proc_info_to_tr(item, row_id, pid))
        elif row_id < cache_length:
            if item["status"] != "dead":
                cache_item = Taska.LATEST_PROC_CACHE[pid]
                cache_item["status"] = "dead"
                cache_item["start_at"], cache_item["end_at"], cache_item["elapsed"] = (
                    Taska.get_end_at(item)
                )
                item = cache_item
            tr_list.append(proc_info_to_tr(item, row_id, pid))
        else:
            Taska.LATEST_PROC_CACHE.pop(pid, None)
    html = Config.console_template.substitute(
        th_list="\n".join(th_list), tr_list="\n".join(tr_list)
    )
    return html


def proc_info_to_tr(item, row_id, pid):
    grep = "%s%s" % (quote_plus('"pid": '), item["pid"])
    href = f'<a target="_blank" href="/view/{item["job_dir"]}">{item["job_dir"]}</a>; <a target="_blank" href="/view/{item["job_dir"]}/result.jsonl?action=view&grep={grep}">result</a>'
    if item["status"] == "running":
        buttons = f"""<td><button onclick='redirect("?kill={pid}&signal=2")'>kill</button></td><td><button onclick='redirect("?kill={pid}&signal=15")'>kill</button></td><td><button onclick='redirect("?kill={pid}&signal=9")' style='color:red'>kill</button></td>"""
    else:
        buttons = """<td>-</td><td>-</td><td>-</td>"""
    return f"""<tr class="{item['status']}"><td>{row_id}</td><td>{item['pid']}</td><td>{item['status']}</td><td>{item['start_at']}</td><td>{item.get('end_at', '-')}</td><td>{item['elapsed']}</td><td>{item['memory']}</td><td>{href}</td>{buttons}</tr>"""


def handle_signal(sig, b):
    logger.warn(f"signal received {sig}, close app")
    sys.exit()


def main(root_path, host="127.0.0.1", port=8021, debug=False):
    Config.root_path = Path(root_path).resolve()
    app.install(AuthPlugin())
    logger.warning(
        f"start server: root_path: {Config.root_path}, debug={debug}, http://{host}:{port}"
    )
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)
    app.run(server="waitress", host=host, port=port, debug=debug)



import logging
from pathlib import Path

from morebuiltins.functools import SizedTimedRotatingFileHandler


class Config:
    LOG_STREAM = True
    LOG_DIR = None
    LOG_LEVEL = logging.DEBUG

    @classmethod
    def init_logger(cls):
        logger = logging.getLogger("taska")
        if not logger.hasHandlers():
            logger.setLevel(cls.LOG_LEVEL)
            fmt = logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")
            if cls.LOG_STREAM:
                stream_handler = logging.StreamHandler()
                stream_handler.setLevel(cls.LOG_LEVEL)
                stream_handler.setFormatter(fmt)
                logger.addHandler(stream_handler)
            if cls.LOG_DIR:
                path = Path(cls.LOG_DIR).joinpath("app.log").resolve()
                path.parent.mkdir(parents=True, exist_ok=True)
                file_handler = SizedTimedRotatingFileHandler(
                    path.as_posix(),
                    when="midnight",
                    interval=1,
                    backupCount=7,
                    maxBytes=50 * 1024 * 1024,
                )
                file_handler.setLevel(cls.LOG_LEVEL)
                file_handler.setFormatter(fmt)
                logger.addHandler(file_handler)
        return logger



console_template = r"""<!DOCTYPE html>
<html lang="en">

<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Console</title>
    <style>
        body {
            font-family: Arial, sans-serif;
        }

        .styled-table {
            width: 100%;
            border-collapse: collapse;
            margin: 25px 0;
            font-size: 0.9em;
            min-width: 400px;
            box-shadow: 0 0 20px rgba(0, 0, 0, 0.15);
        }

        .styled-table thead tr {
            background-color: #009879;
            color: #ffffff;
            text-align: left;
        }

        .styled-table th,
        .styled-table td {
            padding: 12px 15px;
        }

        .styled-table tbody tr {
            border-bottom: 1px solid #dddddd;
        }

        .styled-table tbody tr:nth-of-type(even) {
            background-color: #f3f3f3;
        }

        .styled-table tbody tr:last-of-type {
            border-bottom: 2px solid #009879;
        }
        .dead{
            color: gray;
        }
    </style>
    <script>
        function redirect(url) {
            window.location.href = url;
        }
    </script>
</head>

<body>
    <table class="styled-table">
        <thead>
            <tr>
                $th_list
            </tr>
        </thead>
        <tbody>
            $tr_list
        </tbody>
    </table>

</body>

</html>"""



from argparse import ArgumentParser
from pathlib import Path

from taska.config import Config
from taska.core import Taska


def start_bottle_app(root_path, host="127.0.0.1", port=8021, debug=False):
    from threading import Thread

    from taska.bottle_app.app import main

    taska = Taska()
    thread = Thread(target=taska.run_forever)
    thread.start()
    try:
        return main(root_path, host=host, port=port, debug=debug)
    finally:
        Taska.SHUTDOWN = True
        thread.join()


def main():
    parser = ArgumentParser()
    parser.add_argument("--root", default="", dest="root")
    parser.add_argument("--no-stream-log", action="store_false", dest="stream_log")
    parser.add_argument("--rm-dir", default="", dest="rm_dir")
    parser.add_argument("--launch-job", default="", dest="launch_job")
    parser.add_argument("--ignore-default", action="store_true", dest="ignore_default")
    parser.add_argument(
        "-a",
        "-app",
        "--app",
        "--app-handler",
        default="bottle",
        dest="app_handler",
        help="default/bottle",
    )
    parser.add_argument("--host", default="127.0.0.1", dest="host")
    parser.add_argument("--port", default=8021, type=int, dest="port")
    parser.add_argument("--debug", action="store_true", dest="debug")
    args, extra = parser.parse_known_args()
    if args.root:
        root_path = Path(args.root).resolve()
    elif extra:
        assert len(extra) == 1
        root_path = Path(extra[0]).resolve()
    else:
        raise ValueError("--root is required")
    Taska.ROOT_PATH = root_path
    Config.LOG_STREAM = args.stream_log
    Config.LOG_DIR = root_path.joinpath("logs")
    Config.init_logger()
    if args.rm_dir:
        return print(
            "Removed:", Taska.safe_rm_dir(args.rm_dir), args.rm_dir, flush=True
        )
    elif args.launch_job:
        return Taska.launch_job(Path(args.launch_job))
    else:
        if not args.ignore_default:
            Taska.prepare_default_env(root_path)
        # run app
        if args.app_handler == "default":
            return Taska().run_forever()
        elif args.app_handler == "bottle":
            return start_bottle_app(root_path, args.host, args.port, args.debug)
        elif args.app_handler == "fastapi":
            raise NotImplementedError("fastapi not implemented yet")

        else:
            raise ValueError("--app-handler is required")


if __name__ == "__main__":
    main()



import json
import logging
import os
import re
import signal
import sys
import time
import traceback
import typing
from concurrent.futures import Future
from functools import partial
from logging.handlers import RotatingFileHandler
from pathlib import Path
from threading import Thread, Timer


class LoggerStream:
    def __init__(self, logger):
        self.logger = logger
        self.linebuf = ""
        self.newline = True

    @classmethod
    def setup(cls, std_type: str, dir_path: Path, stdout_limit):
        logger = logging.getLogger(f"{std_type}_log")
        logger.setLevel(logging.DEBUG)
        if not logger.hasHandlers():
            handler = RotatingFileHandler(
                dir_path.joinpath(f"{std_type}.log").resolve().as_posix(),
                maxBytes=stdout_limit * 1.1,
                backupCount=1,
                encoding="utf-8",
                errors="replace",
            )
            handler.terminator = ""
            handler.setLevel(logging.DEBUG)
            formatter = logging.Formatter("%(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        setattr(sys, std_type, LoggerStream(logger))

    def write(self, buf):
        if self.newline:
            buf = f'[{time.strftime("%Y-%m-%d %H:%M:%S")}] {buf}'
        if buf.endswith("\n"):
            self.newline = True
        else:
            self.newline = False
        self.logger.info(buf)

    def flush(self):
        pass


class SingletonError(RuntimeError):
    pass


class MaxWorkersError(RuntimeError):
    pass


class KillError(RuntimeError):
    pass


def handle_signal(sig, frame, future: typing.Optional[Future] = None):
    print(f"[ERROR] Got sig: {sig}, pid: {os.getpid}", flush=True, file=sys.stderr)
    if future:
        future.set_exception(KillError("killed-%s" % sig))


def read_size(text: str):
    # 1g, 1GB, 1g => 1024**3
    m = re.match(r"(\d+)([gGmMkK])?", str(text))
    if not m:
        raise ValueError("Invalid size string: %s" % text)
    a, b = m.groups()
    size = int(a)
    if b:
        size = size * 1024 ** {"g": 3, "m": 2, "k": 1}[b.lower()]
    return size


def is_running(pid: int):
    if sys.platform == "win32":
        with os.popen('tasklist /fo csv /fi "pid eq %s"' % pid) as f:
            f.readline()
            text = f.readline().strip()
            if text:
                return pid
    else:
        try:
            os.kill(pid, 0)
            return pid
        except OSError:
            pass
        except SystemError:
            return pid


def ensure_singleton(current_pid, pid_file: Path):
    if not pid_file.is_file():
        return
    old_pid = pid_file.read_text()
    if not old_pid:
        return
    if is_running(int(old_pid)):
        raise SingletonError(f"Job already running, running_pid: {old_pid}")


def log_result(result_limit, result_item: dict, start_ts):
    result_item["end_at"] = time.strftime("%Y-%m-%d %H:%M:%S")
    result_item["duration"] = round(time.time() - start_ts, 3)
    result_logger = logging.getLogger("result_logger")
    result_logger.setLevel(logging.DEBUG)
    handler = RotatingFileHandler(
        Path(os.getcwd()).joinpath("result.jsonl").resolve().as_posix(),
        maxBytes=result_limit * 1.1,
        backupCount=1,
        encoding="utf-8",
        errors="replace",
    )
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(message)s")
    handler.setFormatter(formatter)
    result_logger.addHandler(handler)
    result_logger.info(json.dumps(result_item, ensure_ascii=False, default=repr))
    handler.flush()
    result_logger.removeHandler(handler)


def setup_stdout_logger(cwd_path, stdout_limit):
    LoggerStream.setup("stdout", cwd_path, stdout_limit)
    LoggerStream.setup("stderr", cwd_path, stdout_limit)


def setup_mem_limit(mem_limit: str):
    if mem_limit:
        if sys.platform == "win32":
            raise ValueError(f"mem_limit={mem_limit} is not supported on windows")
        mem_limit = read_size(mem_limit)
        if mem_limit:
            import resource

            resource.setrlimit(resource.RLIMIT_RSS, (mem_limit, mem_limit))


def start_job(entrypoint, params, workspace_dir, EXEC_GLOBAL_FUTURE: Future):
    pattern = r"^\w+(\.\w+)?(:\w+)?$"
    if re.match(pattern, entrypoint):
        module, _, function = entrypoint.partition(":")
        if module:
            # main may be: 'module.py:main' or 'module.main' or 'package.module:main'
            # replace module.py to module
            sys.path.insert(0, workspace_dir.as_posix())
            module_path = workspace_dir / module
            if module_path.is_file():
                module = module_path.stem
            code = f"import {module}"
            if isinstance(params, dict):
                KWS = params
                ARGS = []
            elif isinstance(params, list):
                KWS = {}
                ARGS = params
            else:
                raise TypeError("Invalid params type: %s. only support list/dict" % type(params))
            if function:
                code += f"; EXEC_GLOBAL_FUTURE.set_result({module}.{function}(*ARGS, **KWS))"
            else:
                code += "; EXEC_GLOBAL_FUTURE.set_result('no result')"
            try:
                exec(
                    code,
                    {
                        "EXEC_GLOBAL_FUTURE": EXEC_GLOBAL_FUTURE,
                        "ARGS": ARGS,
                        "KWS": KWS,
                    },
                )
                if not EXEC_GLOBAL_FUTURE.done():
                    EXEC_GLOBAL_FUTURE.set_exception(
                        RuntimeError("Unknown Error, but no result(not None)")
                    )
            except Exception as e:
                if not EXEC_GLOBAL_FUTURE.done():
                    EXEC_GLOBAL_FUTURE.set_exception(e)

        else:
            EXEC_GLOBAL_FUTURE.set_exception(
                ValueError("Invalid entrypoint: %s" % entrypoint)
            )
    else:
        EXEC_GLOBAL_FUTURE.set_exception(
            ValueError("Invalid entrypoint: %s" % entrypoint)
        )


def ensure_max_workers(root_dir: Path):
    max_workers = int(root_dir.joinpath("max_workers").read_text())
    if max_workers > 0:
        runnings = 0
        for path in root_dir.joinpath("pids").iterdir():
            pid = int(path.name)
            if is_running(pid):
                runnings += 1
        if runnings >= max_workers:
            raise MaxWorkersError(f"Runnings: {runnings}/{max_workers}")


def main():
    """job_meta:
    {
        "entrypoint": "",
        "params": {},
        "enable": 0,
        "crontab": "",
        "timeout": 0,
        "mem_limit": "",
        "result_limit": "",
        "stdout_limit": ""
    }"""
    cwd_path = Path(os.getcwd()).resolve()
    workspace_dir = cwd_path.parent.parent
    root_dir = workspace_dir.parent.parent.parent.parent
    meta = json.loads(cwd_path.joinpath("meta.json").read_text(encoding="utf-8"))
    default_log_size = 5 * 1024**2
    result_limit = read_size(meta["result_limit"] or default_log_size)
    stdout_limit = read_size(meta["stdout_limit"] or default_log_size)
    start_at = time.strftime("%Y-%m-%d %H:%M:%S")
    start_ts = time.time()
    pid = os.getpid()
    result_item = {
        "start_at": start_at,
        "end_at": None,
        "duration": None,
        "pid": pid,
        "result": None,
        "error": None,
    }
    pid_str = str(pid)
    pid_file = cwd_path / "pid.txt"
    global_pid_file = root_dir / "pids" / pid_str
    global_pid_file.parent.mkdir(parents=True, exist_ok=True)
    try:
        thread = None
        ensure_singleton(pid_str, pid_file)
        ensure_max_workers(root_dir)
        global_pid_file.touch()
        # start job
        pid_file.write_text(pid_str)
        cwd_path.joinpath("result.jsonl").touch()
        setup_stdout_logger(cwd_path, stdout_limit)
        EXEC_GLOBAL_FUTURE: Future = Future()
        print(f"[INFO] Job start. pid: {pid_str}", flush=True, file=sys.stderr)
        signal.signal(signal.SIGINT, partial(handle_signal, future=EXEC_GLOBAL_FUTURE))
        signal.signal(signal.SIGTERM, partial(handle_signal, future=EXEC_GLOBAL_FUTURE))
        setup_mem_limit(meta["mem_limit"])
        thread = Thread(
            target=start_job,
            args=(
                meta["entrypoint"],
                meta["params"],
                workspace_dir,
                EXEC_GLOBAL_FUTURE,
            ),
            daemon=True,
        )
        thread.start()

        try:
            timeout = meta.get("timeout")
            if not timeout and isinstance(timeout, int):
                timeout = None
            result_item["result"] = EXEC_GLOBAL_FUTURE.result(timeout=timeout)
        except TimeoutError:
            e = TimeoutError(f"timeout={timeout}")
            if not EXEC_GLOBAL_FUTURE.done():
                EXEC_GLOBAL_FUTURE.set_exception(e)
            raise e
    except Exception as e:
        print(
            f"[ERROR] Job fail. pid: {pid_str}, start_at: {start_at}, error: {traceback.format_exc()}",
            flush=True,
            file=sys.stderr,
        )
        result_item["error"] = repr(e)
    finally:
        log_result(result_limit, result_item, start_ts)
        print(
            f"[INFO] Job end. pid: {pid_str}, start_at: {start_at}",
            flush=True,
            file=sys.stderr,
        )
        if pid_file.is_file() and pid_file.read_text() == pid_str:
            pid_file.unlink(missing_ok=True)
        global_pid_file.unlink(missing_ok=True)
        if thread and thread.is_alive():
            timer = Timer(1, lambda: os._exit(1))
            timer.daemon = True
            timer.start()


if __name__ == "__main__":
    main()



import abc
import json
import logging
import os
import re
import shutil
import signal
import subprocess
import sys
import time
import typing
import venv
from datetime import datetime, timedelta
from hashlib import md5
from pathlib import Path

from morebuiltins.date import Crontab
from morebuiltins.utils import is_running, ptime, read_size, read_time, ttime
from psutil import NoSuchProcess, Process

logger = logging.getLogger("taska")


class Job(typing.TypedDict):
    name: str
    description: str
    # entrypoint = 'package.module:function'
    entrypoint: str
    # params = { 'key': 'value' }
    params: dict
    # enable = 1, 0; 1 = enable, 0 = disable
    enable: int
    # crontab default = 0, format = '* * * * *'
    crontab: str
    # default = 60s
    timeout: int
    # 1g/1gb/1GB == 1024**3
    mem_limit: str
    result_limit: str
    stdout_limit: str


class DirBase(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def prepare_dir(
        cls, target_dir: Path, name: str = "", force=False, **kwargs
    ) -> Path:
        raise NotImplementedError()

    @classmethod
    def is_valid(cls, path: Path):
        raise NotImplementedError()

    @classmethod
    def get_dir_type(cls, path: Path, default="Dir"):
        for d in [RootDir, PythonDir, VenvDir, WorkspaceDir, JobDir]:
            if d.is_valid(path):
                return d.__name__.replace("Dir", "")
        return default


class RootDir(DirBase):
    @classmethod
    def prepare_dir(cls, target_dir: Path, name: str = "root", force=False, **kwargs):
        root_dir = target_dir / name
        if not force and cls.is_valid(root_dir):
            return root_dir.resolve()
        logger.info(f"[Init] Creating root_dir: {root_dir.resolve().as_posix()}")
        root_dir.mkdir(parents=True, exist_ok=True)
        root_dir.joinpath("pids").mkdir(parents=True, exist_ok=True)
        runner_code = Path(__file__).parent.joinpath("templates/runner.py").read_bytes()
        root_dir.joinpath("runner.py").write_bytes(runner_code)
        root_dir.joinpath("max_workers").write_text(
            str(os.cpu_count()), encoding="utf-8"
        )
        assert cls.is_valid(root_dir)
        return root_dir.resolve()

    @classmethod
    def is_valid(cls, path: Path):
        for name in ("runner.py", "pids", "max_workers"):
            if not path.joinpath(name).exists():
                return False
        return True


class PythonDir(DirBase):
    @classmethod
    def prepare_dir(cls, target_dir: Path, name: str = "", force=False, **kwargs):
        python_dir = target_dir / name
        if not force and cls.is_valid(python_dir):
            return python_dir.resolve()
        logger.info(f"[Init] Creating python_dir: {python_dir.resolve().as_posix()}")
        python: str = kwargs.get("python") or sys.executable
        python_path = Path(python)
        if not python_path.exists():
            raise FileNotFoundError(str(python))
        if not name:
            name = re.sub(
                r'[<>:"/\\|?*]', "_", python_path.with_suffix("").resolve().as_posix()
            )
        python_dir.mkdir(parents=True, exist_ok=True)
        python_dir.joinpath("python_path").write_text(
            python_path.resolve().as_posix(), encoding="utf-8"
        )
        assert cls.is_valid(python_dir)
        return python_dir.resolve()

    @classmethod
    def is_valid(cls, path: Path):
        return path.joinpath("python_path").is_file()


class VenvDir(DirBase):
    @classmethod
    def prepare_dir(cls, target_dir: Path, name="", force=False, **kwargs):
        venv_dir = target_dir / name
        if not force and cls.is_valid(venv_dir):
            cls.ensure_pip_install(venv_dir)
            return venv_dir.resolve()
        logger.info(f"[Init] Creating venv_dir: {venv_dir.resolve().as_posix()}")

        builder = venv.EnvBuilder(
            system_site_packages=False,
            clear=True,
            symlinks=False,
            upgrade=False,
            with_pip=True,
            prompt=None,
            upgrade_deps=False,
        )
        builder.create(venv_dir.resolve().as_posix())
        pips = kwargs.get("pips") or []
        venv_dir.joinpath("requirements.txt").write_text("\n".join(pips))
        # PIP_CONFIG_FILE
        venv_dir.joinpath("pip.conf").write_text(
            "[global]\ntimeout = 60", encoding="utf-8"
        )
        venv_dir.joinpath("workspaces").mkdir()
        cls.ensure_pip_install(venv_dir)
        assert cls.is_valid(venv_dir)
        return venv_dir.resolve()

    @classmethod
    def pip_install(cls, venv_dir: Path):
        req_file = venv_dir / "requirements.txt"
        if sys.platform == "win32":
            executable = (venv_dir / "Scripts" / "python.exe").resolve().as_posix()
        else:
            executable = (venv_dir / "bin" / "python").resolve().as_posix()
        cmd = [
            executable,
            "-m",
            "pip",
            "install",
            "-r",
            req_file.resolve().as_posix(),
        ]
        env = os.environ.copy()
        env["PIP_CONFIG_FILE"] = venv_dir.joinpath("pip.conf").resolve().as_posix()
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf-8",
            errors="replace",
            env=env,
        )
        if proc.returncode != 0:
            raise RuntimeError(f"pip install failed {proc.stderr}")
        logger.debug(f"[PIP] {cmd}\n{proc.stdout}")
        return True

    @classmethod
    def ensure_pip_install(cls, path: Path):
        r = path / "requirements.txt"
        m = path / "requirements.md5"
        old_md5 = m.read_text(encoding="utf-8") if m.is_file() else ""
        current_md5 = md5(r.read_bytes()).hexdigest()
        ok = False
        if old_md5 != current_md5:
            m.unlink(missing_ok=True)
            if r.is_file():
                requirements = r.read_text(encoding="utf-8")
            else:
                requirements = ""
            if requirements:
                req_str = requirements.replace("\n", ", ")
                logger.info(f"[Pip] Pip install: {r.resolve().as_posix()} {req_str}")
                ok = cls.pip_install(path)
            else:
                ok = True
            if ok:
                m.write_text(current_md5, encoding="utf-8")
        return str(ok)

    @classmethod
    def is_valid(cls, path: Path):
        return (
            path.joinpath("requirements.txt").is_file()
            and path.joinpath("pip.conf").is_file()
        )


class WorkspaceDir(DirBase):
    @classmethod
    def prepare_dir(cls, target_dir: Path, name: str = "", force=False, **kwargs):
        workspace_dir = target_dir / name
        if not force and cls.is_valid(workspace_dir):
            return workspace_dir.resolve()
        logger.info(
            f"[Init] Creating workspace_dir: {workspace_dir.resolve().as_posix()}"
        )
        workspace_dir.mkdir(parents=True, exist_ok=True)
        workspace_dir.joinpath("jobs").mkdir(parents=True, exist_ok=True)
        assert cls.is_valid(workspace_dir)
        return workspace_dir.resolve()

    @classmethod
    def is_valid(cls, path: Path):
        return path.joinpath("jobs").is_dir()


class JobDir(DirBase):
    default_meta = {
        "name": "job_name",
        "description": "",
        "entrypoint": "",
        "params": {},
        "enable": 0,
        "crontab": "* * * * *",
        "timeout": 0,
        "mem_limit": "",
        "result_limit": "",
        "stdout_limit": "",
    }

    @classmethod
    def prepare_dir(cls, target_dir: Path, name: str = "", force=False, **kwargs):
        job_dir = target_dir / name
        if not force and cls.is_valid(job_dir):
            return job_dir.resolve()
        logger.info(f"[Init] Creating job_dir: {job_dir.resolve().as_posix()}")
        job_dir.mkdir(parents=True, exist_ok=True)
        meta = dict(cls.default_meta)
        meta["name"] = name
        job_dir.joinpath("meta.json").write_text(
            json.dumps(meta, indent=2, ensure_ascii=False), encoding="utf-8"
        )
        assert cls.is_valid(job_dir)
        return job_dir.resolve()

    @classmethod
    def is_valid(cls, path: Path):
        return path.joinpath("meta.json").is_file()


class Taska:
    SHUTDOWN = False
    ROOT_PATH: typing.Optional[Path] = None
    TREE_LEVELS = [RootDir, PythonDir, VenvDir, WorkspaceDir, JobDir]
    LATEST_PROC_CACHE: dict = {}
    CACHE_LENGTH = 50

    def __init__(self):
        if self.ROOT_PATH is None:
            raise ValueError("Taska.ROOT_PATH is not set")
        self.root_dir = self.ROOT_PATH
        self.tree = self.init_dir_tree()
        signal.signal(signalnum=signal.SIGINT, handler=self.handle_shutdown)
        signal.signal(signalnum=signal.SIGTERM, handler=self.handle_shutdown)

    def run_forever(self):
        current_min = time.strftime("%M")
        next_min = (datetime.now() + timedelta(minutes=1)).replace(
            second=0, microsecond=0
        )
        logger.warning(
            f"[Start] Program start, pid={os.getpid()}, root_dir={self.root_dir.resolve().as_posix()}"
        )
        while not self.SHUTDOWN:
            _min = time.strftime("%M")
            if _min != current_min:
                current_min = _min
                self.run_once()
                next_min = (datetime.now() + timedelta(minutes=1)).replace(
                    second=0, microsecond=0
                )
            timeleft = next_min.timestamp() - time.time()
            interval = min((1, timeleft))
            if interval > 0:
                time.sleep(interval)
        logger.warning("[End] Program shutdown")

    def run_once(self):
        for job, path in self.get_todos():
            logger.info(
                f"[Launch] Launch job `{job['name']}`: {path.resolve().as_posix()}"
            )
            self.launch_job(path)

    def handle_shutdown(self, *args):
        self.__class__.SHUTDOWN = True
        logger.warning(f"[Shutdown] received shutdown signal: {args[0]}")

    def need_run(self, now, cron):
        for _ in Crontab.iter_datetimes(cron, start_date=now, max_tries=1):
            return True
        return False

    def get_todos(
        self, now: typing.Optional[datetime] = None
    ) -> typing.Iterator[typing.Tuple[Job, Path]]:
        now = now or datetime.now()
        for path in self.root_dir.rglob("meta.json"):
            job = json.loads(path.read_text(encoding="utf-8"))
            if job["enable"] and job["crontab"] and self.need_run(now, job["crontab"]):
                yield job, path

    def init_dir_tree(self):
        result = {}
        for d in self.root_dir.iterdir():
            if PythonDir.is_valid(d):
                py_result = result.setdefault(d.name, {})
                for v in d.iterdir():
                    if VenvDir.is_valid(v):
                        venv_result = py_result.setdefault(v.name, {})
                        for w in v.joinpath("workspaces").iterdir():
                            if WorkspaceDir.is_valid(w):
                                w_result = venv_result.setdefault(w.name, {})
                                for j in w.joinpath("jobs").iterdir():
                                    if JobDir.is_valid(j):
                                        w_result[j.name] = None
        return result

    @classmethod
    def prepare_default_env(cls, root_dir: Path, force=False):
        # 1. prepare root dir
        root_dir = RootDir.prepare_dir(root_dir.parent, name=root_dir.name, force=force)
        # 2. prepare python dir
        python_dir = PythonDir.prepare_dir(
            root_dir, "default", force=force, python=sys.executable
        )
        # 3. prepare venv dir
        venv_dir = VenvDir.prepare_dir(
            python_dir, "venv1", force=force, pips=["morebuiltins"]
        )
        VenvDir.ensure_pip_install(venv_dir)
        # 4. prepare workspace dir
        workspace_dir = WorkspaceDir.prepare_dir(
            venv_dir.joinpath("workspaces"), name="workspace1", force=force
        )
        # 4.1 add code
        code_path = workspace_dir.joinpath("mycode.py")
        if force or not code_path.is_file():
            code_path.parent.mkdir(parents=True, exist_ok=True)
            code_path.write_text(
                """import time\n\ndef main(arg): return print(time.strftime('%Y-%m-%d %H:%M:%S'), arg) or 'result'""",
                encoding="utf-8",
            )
        # 5. prepare job dir
        job_dir = JobDir.prepare_dir(workspace_dir / "jobs", "job1", force=force)
        return root_dir, python_dir, venv_dir, workspace_dir, job_dir

    @classmethod
    def launch_job(cls, job_path_or_dir: typing.Union[Path, str], timeout=0) -> Path:
        job_path = Path(job_path_or_dir).resolve()
        if job_path.is_dir() and job_path.joinpath("meta.json").is_file():
            job_dir = job_path
        elif job_path.is_file() and job_path.parent.joinpath("meta.json").is_file():
            job_dir = job_path.parent
        else:
            raise FileNotFoundError(job_path)
        if not JobDir.is_valid(job_dir):
            raise FileNotFoundError
        # job dir -> job meta file
        job_path = job_dir / "meta.json"
        workspace_dir = job_dir.parent.parent
        venv_dir = workspace_dir.parent.parent
        runner_path = venv_dir.parent.parent / "runner.py"
        assert runner_path.is_file()
        if sys.platform == "win32":
            executable = venv_dir / "Scripts" / "python.exe"
        else:
            executable = venv_dir / "bin" / "python"
        cmd = [executable.as_posix(), runner_path.as_posix()]
        if sys.platform == "win32":
            proc = subprocess.Popen(
                cmd,
                creationflags=subprocess.DETACHED_PROCESS
                | subprocess.CREATE_NEW_PROCESS_GROUP
                | subprocess.CREATE_NO_WINDOW
                | subprocess.CREATE_BREAKAWAY_FROM_JOB,
                cwd=job_dir.as_posix(),
            )
        else:
            proc = subprocess.Popen(cmd, start_new_session=True, cwd=job_dir.as_posix())
        setattr(proc, "_child_created", False)
        pid_path = job_dir / "pid.txt"
        pid = None
        for _ in range((timeout or 1) * 10):
            try:
                try:
                    if not pid:
                        pid = int(pid_path.read_bytes() or 0)
                        items: dict = cls.get_pids_info([pid])
                        cls.LATEST_PROC_CACHE.update(items)
                        jobs = sorted(
                            cls.LATEST_PROC_CACHE.items(),
                            key=lambda x: x[1]["start_at"],
                        )
                        for k, v in jobs:
                            if v["status"] == "running":
                                try:
                                    v["status"] = Process(k).status()
                                except NoSuchProcess:
                                    v["status"] = "dead"
                            elif len(cls.LATEST_PROC_CACHE) > cls.CACHE_LENGTH:
                                cls.LATEST_PROC_CACHE.pop(k, None)
                except FileNotFoundError:
                    pass
                if proc.wait(0.1):
                    if pid:
                        item = Taska.LATEST_PROC_CACHE[pid]
                        item["status"] = "dead"
                        item["start_at"], item["end_at"], item["elapsed"] = (
                            Taska.get_end_at(item)
                        )
                    break
            except subprocess.TimeoutExpired:
                pass

        del proc
        return job_dir

    @classmethod
    def safe_rm_dir(cls, path: typing.Union[Path, str]):
        path = Path(path)
        if not path.is_dir():
            return True
        for pid_path in path.rglob("pid.txt"):
            try:
                pid = pid_path.read_bytes()
                if is_running(pid):
                    return False
            except FileNotFoundError:
                continue
        shutil.rmtree(path.resolve().as_posix(), ignore_errors=True)
        return not path.is_dir()

    @classmethod
    def get_pids_info(cls, pids: typing.List[int]):
        items: dict = {}
        root_path = cls.ROOT_PATH
        if root_path is None:
            raise RuntimeError("Taska.ROOT_PATH is not set")
        now = time.time()
        for pid in pids:
            try:
                proc = Process(pid)
                status = proc.status()
                job_dir = Path(proc.cwd()).relative_to(root_path).as_posix()
                start_at = ttime(proc.create_time())
                elapsed = read_time(now - proc.create_time(), shorten=True)
                memory = read_size(proc.memory_info().rss, shorten=True)
            except NoSuchProcess:
                proc = None
                job_dir = status = start_at = elapsed = memory = "-"
            item = {
                "pid": pid,
                "status": status,
                "job_dir": job_dir,
                "start_at": start_at,
                "elapsed": elapsed,
                "memory": memory,
            }
            items[pid] = item
        return items

    @classmethod
    def get_end_at(cls, item):
        if not cls.ROOT_PATH:
            raise ValueError("Taska.ROOT_PATH is not set")
        result_path = cls.ROOT_PATH.joinpath(item["job_dir"]).joinpath("result.jsonl")
        result = (item["start_at"], "-", "-")
        if result_path.is_file():
            with open(result_path, "r", encoding="utf-8") as f:
                pid = item["pid"]
                regex = re.compile(f'"pid": ?{pid},')
                for line in f:
                    if regex.search(line):
                        data = json.loads(line)
                        if data.get("pid") == pid:
                            result = (
                                data["start_at"],
                                data["end_at"],
                                read_time(
                                    ptime(data["end_at"]) - ptime(data["start_at"]),
                                    shorten=True,
                                ),
                            )
        return result


def test():
    Taska.ROOT_PATH = Path("../demo_path/").resolve()
    Taska.prepare_default_env(Taska.ROOT_PATH, force=False)
    ta = Taska()
    print(ta.tree)
    print(list(ta.get_todos()))
    print(datetime.now())
    for job, path in ta.get_todos():
        print(job, "launch")
        ta.launch_job(path)


if __name__ == "__main__":
    test()
